#include <zheaders.h>
#include <string.h>
#include "zau-comm.h"

#define RECV_QUE_SIZE	(4096)
#define SEND_QUE_SIZE	(2048)

#define COND_BR_MASK    0xff00
#define UCOND_DBR_MASK  0xe000
#define UCOND_RBR_MASK  0xf0df
#define TRAPA_MASK      0xff00

#define COND_DISP               0x00ff
#define UCOND_DISP              0x0fff
#define UCOND_REG               0x0f00

#define BF_INSTR                0x8b00
#define BFS_INSTR               0x8f00
#define BT_INSTR                0x8900
#define BTS_INSTR               0x8d00
#define BRA_INSTR               0xa000
#define BSR_INSTR               0xb000
#define BSRF_INSTR              0x0003
#define JMP_INSTR               0x402b
#define JSR_INSTR               0x400b
#define RTS_INSTR               0x000b
#define RTE_INSTR               0x002b
#define TRAPA_INSTR             0xc300

#define SSTEP_INSTR             0xc3ff

#define T_BIT_MASK              0x0001

#define BUFMAX      2048
#define NUMREGBYTES 112
#define NUMREGS     28

enum regnames{
  R0, R1, R2, R3, R4, R5, R6, R7,
  R8, R9, R10, R11, R12, R13, R14,
  R15, PC, PR, GBR, VBR, MACH, MACL, SR,
  TICKS, STALLS, CYCLES, INSTS, PLR
};


#define cputs(s) comm_write(s,strlen(s))

typedef struct
{
  short *memAddr;
  short oldInstr;
}stepData;

static OBJECT_NAME	target_code = 0;
static OBJECT_NAME	target_proc = 0;

BCOM    *pBCom;
COMMDEV *pDev;

stepData instrBuffer;
char     stepped;

BYTE4 registers[NUMREGS];

int      remote_debug;
int      dofault;

int comm_init()
{
  BCOMMODE mode;
  BCOMSTAT State;

  if ((pDev = LockCommDev(BLTIN_15PIN_DEV)) == NULL) {
    return 0;
  }

  if ((pBCom = OpenBComm(pDev, RECV_QUE_SIZE, SEND_QUE_SIZE)) == NULL) {
    UnlockCommDev(pDev);
    return 0;
  }
  
  GetModeBComm(pBCom, &mode);
  
  mode.ByteSize = DATABIT_8;
  mode.Parity = PARITY_NONE;
  mode.StopBits = STOPBIT_1;
  mode.BaudRate = 115200;
  mode.fTxCtsLook = 1;
  mode.fTxDsrLook = 0;
  mode.fRxDtrDrive = 0;
  mode.fRxRtsDrive = 1;
  mode.fRxXSend = 0;
  mode.fTxXLook = 0;
  
  SetModeBComm(pBCom, &mode);

  return 1;
}

int
comm_write(buf, n)
     char *buf;
     int n;
{
  BCOMSTAT State;
  int r;
  
  r = WaitChangeBCommState(pBCom, &State, BCOM_EVENT_VACANT_TXQUE, BCOM_WAIT_FOREVER);
  
  if (r < 0) {
    r = GetBCommError(pBCom);
    return 0;
  }
  
  r = WriteBComm(pBCom, buf, n);
  
  if (r < 0) {
    r = GetBCommError(pBCom);
    return 0;
  }  

  return r;
}

int
comm_read(buf, n)
     char *buf;
     int n;
{
  BCOMSTAT State;
  int r;
  
  r = WaitChangeBCommState(pBCom, &State, BCOM_EVENT_DATA_ARRIVE, BCOM_WAIT_FOREVER);
    
  if (r < 0){
    r = GetBCommError(pBCom);
    return 0;
  }

  r = ReadBComm(pBCom, buf, n);

  if (r < 0) {
    r = GetBCommError(pBCom);
    return 0;
  }

  /*  comm_write(buf,r);*/

  return r;
}

static int
tohex (nib)
     int nib;
{
  if (nib < 10)
    return '0' + nib;
  else
    return 'a' + nib - 10;
}

static int
fromhex (a)
     int a;
{
  if (a >= '0' && a <= '9')
    return a - '0';
  else if (a >= 'a' && a <= 'f')
    return a - 'a' + 10;

}

static int
readchar ()
{
  static char buf[256];
  static int bufcnt = 0;
  static char *bufp;

  if (bufcnt-- > 0)
    return *bufp++ & 0x7f;

  bufcnt = comm_read (buf, sizeof (buf));

  if (bufcnt <= 0)
    {
      return -1;
    }

  bufp = buf;
  bufcnt--;
  return *bufp++ & 0x7f;
}

int
getpkt (buf)
     char *buf;
{
  char *bp;
  unsigned char csum, c1, c2;
  int c;

  while (1)
    {
      csum = 0;

      while (1)
        {
          c = readchar ();
          if (c == '$')
            break;
          if (c < 0)
            return -1;
        }

     bp = buf;
      while (1)
        {
          c = readchar ();
          if (c < 0)
            return -1;
          if (c == '#')
            break;
          *bp++ = c;
          csum += c;
        }
      *bp = 0;

      c1 = fromhex (readchar ());
      c2 = fromhex (readchar ());
      
      if (csum == (c1 << 4) + c2)
        break;

      comm_write ( "-", 1);
    }

  comm_write ( "+", 1);

  return bp - buf;
}

int
putpkt (buf)
     char *buf;
{
  int i;
  unsigned char csum = 0;
  char buf2[2000];
  char buf3[1];
  int cnt = strlen (buf);
  char *p;

  /* Copy the packet into buffer BUF2, encapsulating it
     and giving it a checksum.  */

  p = buf2;
  *p++ = '$';
  
  for (i = 0; i < cnt; i++)
    {
      csum += buf[i];
      *p++ = buf[i];
    }
  *p++ = '#';
  *p++ = tohex ((csum >> 4) & 0xf);
  *p++ = tohex (csum & 0xf);

  *p = '\0';
  
  /* Send it over and over until we get a positive ack.  */

  do
    {
      int cc,r=0;
      
      if (comm_write (buf2, p - buf2) != p - buf2)
        {
          return -1;
        }

      cc = comm_read( buf3, 1);

      if (cc <= 0)
	{
          return -1;
        }

      r++;
      if(r>5) return -1;
    }
  while (buf3[0] != '+');

  return 1;                     /* Success! */
}

char *
mem2hex (char *mem, char *buf, int count)
{
  int i;
  int ch;
  
  for (i = 0; i < count; i++){
    ch=dbg_mem_get1(target_proc,mem+i);

    *buf++ = tohex ((ch >> 4) & 0xf);
    *buf++ = tohex (ch & 0xf);
  }
  *buf = 0;

  return (buf);
}

char *
hex2mem (char *buf, char *mem, int count)
{
  int i;
  int ch;

  for (i = 0; i < count; i++){
    ch = fromhex (*buf++) << 4;
    ch = ch + fromhex(*buf++);
    dbg_mem_put1(target_proc,mem+i,ch);
  }

  return (mem);
}

read_registers()
{
  int i;

  for(i=0;i<16;i++)
    registers[i]=dbg_reg_get(target_proc,i);

  registers[SR]=dbg_reg_get(target_proc,REG_SR);
  registers[GBR]=dbg_reg_get(target_proc,REG_GBR);
  registers[MACH]=dbg_reg_get(target_proc,REG_MACH);
  registers[MACL]=dbg_reg_get(target_proc,REG_MACL);
  registers[PR]=dbg_reg_get(target_proc,REG_PR);
  registers[PC]=dbg_reg_get(target_proc,REG_PC);

  return 1;
}

write_registers()
{
  int i;

  for(i=0;i<16;i++)
    dbg_reg_put(target_proc,i,registers[i]);

  dbg_reg_put(target_proc,REG_SR,registers[SR]);
  dbg_reg_put(target_proc,REG_GBR,registers[GBR]);
  dbg_reg_put(target_proc,REG_MACH,registers[MACH]);
  dbg_reg_put(target_proc,REG_MACL,registers[MACL]);
  dbg_reg_put(target_proc,REG_PR,registers[PR]);
  dbg_reg_put(target_proc,REG_PC,registers[PC]);

  return 1;
}

static char *
reg2hex (char *mem, char *buf, int count)
{
  int i;
  int ch;
  for (i = 0; i < count; i++)
    {
      ch = *mem++;
      *buf++ = tohex ((ch >> 4) & 0xf);
      *buf++ = tohex (ch & 0xf);
    }
  *buf = 0;
  return (buf);
}

static char *
hex2reg (char *buf, char *mem, int count)
{
  int i;
  unsigned char ch;
  for (i = 0; i < count; i++)
    {
      ch = fromhex (*buf++) << 4;
      ch = ch + fromhex(*buf++);
      *mem++ = ch;
    }
  return (mem);
}

int hexToInt (char **ptr, int *intValue)
{
  int numChars = 0;
  int hexValue;
  
  *intValue = 0;
  
  while (**ptr)
    {
      hexValue = fromhex (**ptr);
      if (hexValue >= 0)
        {
          *intValue = (*intValue << 4) | hexValue;
          numChars++;
        }
      else
        break;
      
      (*ptr)++;
    }

  return (numChars);
}

doSStep (void)
{
  short *instrMem;
  int displacement;
  int reg;
  unsigned short opcode;
  unsigned short val;
  char *addr;

  read_registers();
  
  instrMem = (short *) registers[PC];
  addr = (char *) registers[PC];

  /*  opcode = *instrMem;*/
  val=dbg_mem_get1(target_proc,addr+1);
  val=(val << 8)+dbg_mem_get1(target_proc,addr);
  opcode = val;
  
  stepped = 1;
  
  if ((opcode & COND_BR_MASK) == BT_INSTR)
    {
      if (registers[SR] & T_BIT_MASK)
        {
          displacement = (opcode & COND_DISP) << 1;
          if (displacement & 0x100)
            displacement |= 0xfffffe00;
	  instrMem = (short *) (registers[PC] + displacement + 4);
        }
      else
        instrMem += 1;
    }
  else if ((opcode & COND_BR_MASK) == BTS_INSTR)
    {
      if (registers[SR] & T_BIT_MASK)
        {
          displacement = (opcode & COND_DISP) << 1;
          if (displacement & 0x100)
            displacement |= 0xfffffe00;
	  instrMem = (short *) (registers[PC] + displacement + 4);
        }
      else
        instrMem += 2;
    }
  else if ((opcode & COND_BR_MASK) == BF_INSTR)
    {
      if (registers[SR] & T_BIT_MASK)
        instrMem += 1;
      else
        {
          displacement = (opcode & COND_DISP) << 1;
          if (displacement & 0x100)
            displacement |= 0xfffffe00;

	  instrMem = (short *) (registers[PC] + displacement + 4);
        }
    }
  else if ((opcode & COND_BR_MASK) == BFS_INSTR)
    {
      if (registers[SR] & T_BIT_MASK)
        instrMem += 2;
      else
        {
          displacement = (opcode & COND_DISP) << 1;
          if (displacement & 0x100)
            displacement |= 0xfffffe00;

	  instrMem = (short *) (registers[PC] + displacement + 4);
        }
    }
  else if ((opcode & UCOND_DBR_MASK) == BRA_INSTR)
    {
      displacement = (opcode & UCOND_DISP) << 1;
      if (displacement & 0x01000)
        displacement |= 0xffffe000;
      instrMem = (short *) (registers[PC] + displacement + 4);
    }
  else if ((opcode & UCOND_RBR_MASK) == JSR_INSTR)
    {
      reg = (char) ((opcode & UCOND_REG) >> 8);
      instrMem = (short *) registers[reg];
    }
  else if ((opcode & UCOND_RBR_MASK) == BSRF_INSTR)
    {
      reg = (char) ((opcode & UCOND_REG) >> 8);

      instrMem = (short *) (registers[reg] + registers[PC] + 4);
    }
  else if (opcode == RTS_INSTR)
    instrMem = (short *) registers[PR];
  else if (opcode == RTE_INSTR)
    instrMem = (short *) registers[15];
  /*
  else if ((opcode & TRAPA_MASK) == TRAPA_INSTR)
    instrMem = (short *) ((opcode & ~TRAPA_MASK) << 2);
    */
  else
    instrMem += 1;

  /*
    instrBuffer.memAddr = instrMem;
    instrBuffer.oldInstr = *instrMem;
    *instrMem = SSTEP_INSTR;
    */

  instrBuffer.memAddr = instrMem;
  dbg_break_point_set(target_proc,(long)instrMem);

}

void
undoSStep (void)
{
  if (stepped)
    {
      short *instrMem;
      
      instrMem = instrBuffer.memAddr;
      dbg_break_point_clear(target_proc,instrMem);

      /*      *instrMem = instrBuffer.oldInstr;*/
    }
  stepped = 0;
}

doCommand(in,out)
     char *in,*out;
{
  int  sigval=7,p;
  int  addr,length;
  char *ptr;

  switch (in[0]){
  case '?':
    out[0] = 'S';
    out[1] = tohex ((sigval >> 4)&0xf);
    out[2] = tohex (sigval & 0xf);
    out[3] = 0;
    break;

  case 'd':
    remote_debug = !(remote_debug);
    break;
    
  case 'g':
    read_registers();
    reg2hex ((char *)registers,out, NUMREGBYTES);
    break;
  case 'G':
    hex2reg (&in[1],(char *)registers, NUMREGBYTES);
    write_registers();
    strcpy (out, "OK");
    break;
    
  case 'm':
    dofault = 0;
    
    ptr = &in[1];
    if (hexToInt (&ptr, &addr))
      if (*(ptr++) == ',')
	if (hexToInt (&ptr, &length)){
	  ptr = 0;
	  mem2hex ((char *) addr, out, length);
	}
    if (ptr)
      strcpy (out, "E01");
    
    dofault = 1;
    break;
    
  case 'M':
    dofault = 0;
    
    ptr = &in[1];
    if (hexToInt (&ptr, &addr))
      if (*(ptr++) == ',')
	if (hexToInt (&ptr, &length))
	  if (*(ptr++) == ':'){
	    hex2mem (ptr, (char *) addr, length);
	    ptr = 0;
	    strcpy (in, "OK");
	  }
    if (ptr)
      strcpy (out, "E02");
    
    dofault = 1;
	break;
	
  case 's':
  case 'c':
    ptr = &in[1];
    if (hexToInt (&ptr, &addr)){
      read_registers();	/* 2000/Mar/14 */
      registers[PC] = addr;
      write_registers();
    }
    
    if(in[0]=='s') doSStep();
    
    dbg_proc_resume(target_proc);
    
    while(dbg_is_proc_breaked(target_proc)==0){
      BCOMSTAT State;	
      int  r;
      char c=0;
      
      r = WaitChangeBCommState(pBCom, &State, BCOM_EVENT_DATA_ARRIVE, BCOM_WAIT_1SEC);
      
      if (r >= 0){
	ReadBComm(pBCom, &c, 1);
	if(c=='\003'){
	  if(dbg_is_proc_breaked(target_proc)==0){
	    dbg_proc_stop(target_proc);
	    sigval = 2;
	    goto abort;
	  }
	}
      }
    }
    
    p=dbg_is_proc_breaked(target_proc);
    
    if(p==-1) sigval = 3;
    else      sigval = 5;
    
  abort:
    undoSStep();
    
    out[0] = 'S';
    out[1] = tohex ((sigval >> 4)&0xf);
    out[2] = tohex (sigval & 0xf);
    out[3] = 0;
    
    break;
    
  case 'k':
    if(target_proc){
      proc_stop(target_proc);
      proc_delete(target_proc);
    }
    target_proc=0;
    break;
    
  case 'Z':
    ptr = &in[1];
    if (hexToInt (&ptr, &addr)){
      dbg_break_point_set(target_proc,(short *)addr);
      strcpy(out,"OK");
    }else{
      strcpy(out,"ENN");
    }
    break;
    
  case 'z':
    ptr = &in[1];
    if (hexToInt (&ptr, &addr)){
      dbg_break_point_clear(target_proc,(short *)addr);
      strcpy(out,"OK");
    }else{
      strcpy(out,"ENN");
    }
    break;
    
  default:
    out[0]='\0';
    break;
  }        

}

int chksio(buf)
     char *buf;
{
  int  i,j,r;
  char c;
  BCOMSTAT State;

  r = WaitChangeBCommState(pBCom, &State, BCOM_EVENT_VACANT_TXQUE, BCOM_WAIT_1SEC);
  
  if(r<0){
    r=GetBCommError(pBCom);
    return 0;
  }

  i=ReadBComm(pBCom, &c, 1);

  /*i=comm_read(&c,1);*/

  if(i<0){
    r=GetBCommError(pBCom);
    return 0;
  }
  
  if(i==0) return 0;

  for(j=0;j<i;j++){
    r=getpkt_t(c,buf);
    if(r) return r;
  }

  return 0;
}

int getpkt_t(c,buf)
     char c;
     char *buf;
{
  static int stat;
  static int ptr=0;
  static unsigned char  csum,chk;
  /*
  {
    char s[256];
    sprintf(s,"stat:%d\n",stat);
    cputs(s);
  }
  */
  switch(stat){
  case 0:
    if(c!='$') return 0;
    csum=0;
    stat=1;
    break;

  case 1:
    if(c=='#'){
      stat=2;
      buf[ptr]='\0';
      ptr=0;
    }else{
      buf[ptr]=c;
      csum += c;
      ptr++;
    }
    break;

  case 2:
    chk = fromhex (c);
    chk <<=4;
    stat=3;
    break;

  case 3:
    chk += fromhex (c);
    stat=0;

    if (csum == chk){
      comm_write("+",1);
      return 1;
    }else{
      comm_write("-",1);
      return 0;
    }
    break;
  }

  return (0);
}

int
more_main()
{
  PROC_INI	pak;
  /*  OBJECT_NAME	code,proc;*/
  
  OBJECT_NAME	proc;
  char remcomInBuffer[BUFMAX];
  char remcomOutBuffer[BUFMAX];
  int r;

  InitHeap(0);
  comm_init();
  init_queue();

  while (1){
    r=get_target_pname(1,&proc);

    if(!r){
      target_proc=proc;
      read_registers();
    }
    
    /*      getpkt (remcomInBuffer);
	    doCommand(remcomInBuffer,remcomOutBuffer);*/
    
    r=chksio(remcomInBuffer);

    if(r){
      if(target_proc){
	doCommand(remcomInBuffer,remcomOutBuffer);
	putpkt (remcomOutBuffer);
      }
    }
  }

  return 1;
}

memcpy(){}

/* EOF */






